#include <libc64_dsp/include/inc_overlay.h>
//#include "scalers.h"
#include <stdint.h>

#define MAX_WIDTH 320

#define yuv_diff(yuv1, yuv2) ((_cmpgtu4(_subabs4(yuv1, yuv2), 0x00300706))?1:0)


static void prepare_yuv_palette(const uint16_t * restrict RGBpal, uint64_t * restrict RGBYUVpal)
{
    _nassert((int)RGBpal % 8 == 0);
    _nassert((int)RGBYUVpal % 8 == 0);

    unsigned int i;

    for (i = 0; i < 256; i++)
    {
        uint32_t rgbvalue;
        uint32_t r, g, b;
        uint32_t y, u, v;

        rgbvalue = RGBpal[i];
        r = (rgbvalue << 3) & (31 << 3);
        g = (rgbvalue >> 3) & (63 << 2);
        b = (rgbvalue >> 8) & (31 << 3);

        r |= r >> 5;
        g |= g >> 6;
        b |= b >> 5;

        // this is the original HQX formula - using binary floating point numbers
        //y = (uint32_t)(0.299*r + 0.587*g + 0.114*b);
        //u = (uint32_t)(-0.169*r - 0.331*g + 0.5*b) + 128;
        //v = (uint32_t)(0.5*r - 0.419*g - 0.081*b) + 128;

        // this is the same formula, but using more precise coeficients and decimal floating point numbers
        // it gives slightly different results, but it's more precise
        //y = ( 0.299dd   *r + 0.587dd   *g + 0.114dd   *b);
        //u = (-0.168736dd*r - 0.331264dd*g + 0.5dd     *b) + 128.0dd;
        //v = ( 0.5dd     *r - 0.418688dd*g - 0.081312dd*b) + 128.0dd;

        // this is the same formula, but using 9.23 fixed point numbers
        // it gives the results as when using using decimal floating point numbers
        y = ((((uint32_t)( 2508194*r + 4924113*g +  956302*b))             ) >> 23);
        //u = ((((uint32_t)(-1415460*r - 2778844*g + 4194304*b)) + 1073741824) >> 23);
        u = ((((uint32_t)( 4194304*b - 1415460*r - 2778844*g)) + 1073741824) >> 23);
        v = ((((uint32_t)( 4194304*r - 3512210*g -  682094*b)) + 1073741824) >> 23);

        // this is the same formula, but minimizing coeficients and fraction presision
        // it gives the results as when using using decimal floating point numbers
        //
        // this(>>18) works for rgb888 values created from rgb565
        //y = ((((uint32_t)(  78381*r + 153879*g +  29885*b))          ) >> 18);
        // this(>>19) works for all rgb888 values
        //y = ((((uint32_t)( 156763*r + 307758*g +  59769*b))          ) >> 19);
        //u = ((((uint32_t)( -11058*r -  21710*g +  32768*b)) + 8388608) >> 16);
        //v = ((((uint32_t)(  32768*r -  27439*g -   5329*b)) + 8388608) >> 16);

        RGBYUVpal[i] = ((y << 16) + (u << 8) + v) | (((uint64_t)rgbvalue) << 32);
    }
}

static void convert_yuv_rgb_8(const uint8_t * restrict src, uint32_t * restrict yuvdst, uint16_t * restrict rgbdst, const uint64_t * restrict RGBYUVpal, int width)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)yuvdst % 8 == 0);
    _nassert((int)rgbdst % 8 == 0);
    _nassert((int)RGBYUVpal % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int index;

    for (index = 0; index < width; index++)
    {
        uint64_t palvalue;

        palvalue = RGBYUVpal[src[index]];

        yuvdst[index] = _loll(palvalue);
        rgbdst[index] = _hill(palvalue);
    }

    yuvdst[-1] = yuvdst[0];
    yuvdst[width] = yuvdst[width - 1];

    rgbdst[-1] = rgbdst[0];
    rgbdst[width] = rgbdst[width - 1];
}

static void convert_yuv_16(uint16_t * restrict src, uint32_t * restrict dst, int width)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int index;

    src[-1] = src[0];
    src[width] = src[width - 1];

    for (index = 0; index < width; index++)
    {
        uint32_t r, g, b;
        uint32_t y, u, v;

        r = (src[index] << 3) & (31 << 3);
        g = (src[index] >> 3) & (63 << 2);
        b = (src[index] >> 8) & (31 << 3);

        r |= r >> 5;
        g |= g >> 6;
        b |= b >> 5;

        // this is the original HQX formula - using binary floating point numbers
        //y = (uint32_t)(0.299*r + 0.587*g + 0.114*b);
        //u = (uint32_t)(-0.169*r - 0.331*g + 0.5*b) + 128;
        //v = (uint32_t)(0.5*r - 0.419*g - 0.081*b) + 128;

        // this is the same formula, but using more precise coeficients and decimal floating point numbers
        // it gives slightly different results, but it's more precise
        //y = ( 0.299dd   *r + 0.587dd   *g + 0.114dd   *b);
        //u = (-0.168736dd*r - 0.331264dd*g + 0.5dd     *b) + 128.0dd;
        //v = ( 0.5dd     *r - 0.418688dd*g - 0.081312dd*b) + 128.0dd;

        // this is the same formula, but using 9.23 fixed point numbers
        // it gives the results as when using using decimal floating point numbers
        y = ((((uint32_t)( 2508194*r + 4924113*g +  956302*b))             ) >> 23);
        //u = ((((uint32_t)(-1415460*r - 2778844*g + 4194304*b)) + 1073741824) >> 23);
        u = ((((uint32_t)( 4194304*b - 1415460*r - 2778844*g)) + 1073741824) >> 23);
        v = ((((uint32_t)( 4194304*r - 3512210*g -  682094*b)) + 1073741824) >> 23);

        // this is the same formula, but minimizing coeficients and fraction presision
        // it gives the results as when using using decimal floating point numbers
        //
        // this(>>18) works for rgb888 values created from rgb565
        //y = ((((uint32_t)(  78381*r + 153879*g +  29885*b))          ) >> 18);
        // this(>>19) works for all rgb888 values
        //y = ((((uint32_t)( 156763*r + 307758*g +  59769*b))          ) >> 19);
        //u = ((((uint32_t)( -11058*r -  21710*g +  32768*b)) + 8388608) >> 16);
        //v = ((((uint32_t)(  32768*r -  27439*g -   5329*b)) + 8388608) >> 16);

        dst[index] = (y << 16) + (u << 8) + v;
    }

    dst[-1] = dst[0];
    dst[width] = dst[width - 1];
}

static void calculate_pattern(const uint32_t * restrict yuvsrc1, const uint32_t * restrict yuvsrc2, const uint32_t * restrict yuvsrc3, uint8_t * restrict dst, int width)
{
    _nassert((int)yuvsrc1 % 8 == 0);
    _nassert((int)yuvsrc2 % 8 == 0);
    _nassert((int)yuvsrc3 % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    uint32_t e;
    uint8_t pattern;

    // before first
    {
        e = yuvsrc2[-1];
        pattern  = yuv_diff(e, yuvsrc1[-1]) << 0;
        pattern |= yuv_diff(e, yuvsrc1[-1]) << 1;
        pattern |= yuv_diff(e, yuvsrc1[ 0]) << 2;
        pattern |= yuv_diff(e, yuvsrc2[-1]) << 3;
        pattern |= yuv_diff(e, yuvsrc2[ 0]) << 4;
        pattern |= yuv_diff(e, yuvsrc3[-1]) << 5;
        pattern |= yuv_diff(e, yuvsrc3[-1]) << 6;
        pattern |= yuv_diff(e, yuvsrc3[ 0]) << 7;

        dst[-1] = pattern;
    }

    for (; width != 0; width--)
    {
        e = yuvsrc2[0];
        pattern  = yuv_diff(e, yuvsrc1[-1]) << 0;
        pattern |= yuv_diff(e, yuvsrc1[ 0]) << 1;
        pattern |= yuv_diff(e, yuvsrc1[ 1]) << 2;
        pattern |= yuv_diff(e, yuvsrc2[-1]) << 3;
        pattern |= yuv_diff(e, yuvsrc2[ 1]) << 4;
        pattern |= yuv_diff(e, yuvsrc3[-1]) << 5;
        pattern |= yuv_diff(e, yuvsrc3[ 0]) << 6;
        pattern |= yuv_diff(e, yuvsrc3[ 1]) << 7;

        *dst = pattern;

        yuvsrc1++;
        yuvsrc2++;
        yuvsrc3++;
        dst++;
    }

    // after last
    {
        e = yuvsrc2[0];
        pattern  = yuv_diff(e, yuvsrc1[-1]) << 0;
        pattern |= yuv_diff(e, yuvsrc1[ 0]) << 1;
        pattern |= yuv_diff(e, yuvsrc1[ 0]) << 2;
        pattern |= yuv_diff(e, yuvsrc2[-1]) << 3;
        pattern |= yuv_diff(e, yuvsrc2[ 0]) << 4;
        pattern |= yuv_diff(e, yuvsrc3[-1]) << 5;
        pattern |= yuv_diff(e, yuvsrc3[ 0]) << 6;
        pattern |= yuv_diff(e, yuvsrc3[ 0]) << 7;

        *dst = pattern;
    }
}

static void fill_blendtable(const uint8_t * restrict patternTable, const uint32_t * restrict hqTable, const uint8_t * restrict hqTable2, uint16_t * restrict blendTable, int width)
{
    _nassert((int)patternTable % 8 == 0);
    _nassert((int)hqTable % 8 == 0);
    _nassert((int)blendTable % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int index, hqindex;
    uint16_t * restrict blendValues;

    for (index = 0; index < 80; index++)
    {
        blendTable[index] = 0xffff;
    }

    blendValues = (uint16_t *) (blendTable+80);

    _nassert((int)blendValues % 8 == 0);

    // fill blendtables based on patterntable

    for (index = 0; index < width; index++)
    {
        uint32_t hqValue, rule;

        hqValue = hqTable[patternTable[index]];

        rule = hqValue & 0x1f;
        blendValues[index] = blendTable[rule];
        blendTable[rule] = index;

        rule = (hqValue >> 8) & 0x1f;
        blendValues[index + MAX_WIDTH] = blendTable[rule + 20];
        blendTable[rule + 20] = index;

        rule = (hqValue >> 16) & 0x1f;
        blendValues[index + 2*MAX_WIDTH] = blendTable[rule + 2*20];
        blendTable[rule + 2*20] = index;

        rule = (hqValue >> 24) & 0x1f;
        blendValues[index + 3*MAX_WIDTH] = blendTable[rule + 3*20];
        blendTable[rule + 3*20] = index;
    }

    // modify blendtables based on further comparisons

    // upper left pixel

#define blendExpr12 patternTable[index-1] & (1 << 2)
#define blendExpr18 patternTable[index+1] & (1 << 0)
#define blendExpr19 patternTable[index-1] & (1 << 7)

#include "fill_blendtable_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    // upper right pixel

#define blendExpr12 patternTable[index+1] & (1 << 0)
#define blendExpr18 patternTable[index+1] & (1 << 5)
#define blendExpr19 patternTable[index-1] & (1 << 2)

#include "fill_blendtable_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    // lower right pixel

#define blendExpr12 patternTable[index+1] & (1 << 5)
#define blendExpr18 patternTable[index-1] & (1 << 7)
#define blendExpr19 patternTable[index+1] & (1 << 0)

#include "fill_blendtable_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    // lower left pixel

#define blendExpr12 patternTable[index-1] & (1 << 7)
#define blendExpr18 patternTable[index-1] & (1 << 2)
#define blendExpr19 patternTable[index+1] & (1 << 5)

#include "fill_blendtable_proc_body.h"
}

#define GROW(n) n = (n | n << 16) & 0x07e0f81f
static uint16_t inline pack(uint32_t n)
{
    n &= 0x07e0f81f;
    return n | (n >> 16);
}

static uint16_t inline blend1(uint32_t A, uint32_t B)
{
    GROW(A); GROW(B);
    A = (A * 3 + B) >> 2;
    return pack(A);
}

static uint16_t inline blend2(uint32_t A, uint32_t B, uint32_t C)
{
    GROW(A); GROW(B); GROW(C);
    return pack((A * 2 + B + C) >> 2);
}

static uint16_t inline blend3(uint32_t A, uint32_t B, uint32_t C)
{
    GROW(A); GROW(B); GROW(C);
    return pack((A * 5 + B * 2 + C) >> 3);
}

static uint16_t inline blend4(uint32_t A, uint32_t B, uint32_t C)
{
    GROW(A); GROW(B); GROW(C);
    return pack((A * 6 + B + C) >> 3);
}

static uint16_t inline blend5(uint32_t A, uint32_t B, uint32_t C)
{
    GROW(A); GROW(B); GROW(C);
    return pack((A * 2 + (B + C) * 3) >> 3);
}

static uint16_t inline blend6(uint32_t A, uint32_t B, uint32_t C)
{
    GROW(A); GROW(B); GROW(C);
    return pack((A * 14 + B + C) >> 4);
}

static void fill_dst(const uint16_t * restrict rgbsrc1, const uint16_t * restrict rgbsrc2, const uint16_t * restrict rgbsrc3, uint16_t * restrict dst, const uint16_t * restrict blendTable, int width)
{
    _nassert((int)rgbsrc1 % 8 == 0);
    _nassert((int)rgbsrc2 % 8 == 0);
    _nassert((int)rgbsrc3 % 8 == 0);
    _nassert((int)blendTable % 8 == 0);

    int index;
    const uint16_t * restrict blendValues;

    blendValues = (uint16_t *) (blendTable+80);

    _nassert((int)blendValues % 8 == 0);


    _nassert((int)dst % 8 == 0);
    // upper left pixel

#define E rgbsrc2[index]
#define A rgbsrc1[index-1]
#define B rgbsrc1[index]
#define D rgbsrc2[index-1]
#define F rgbsrc2[index+1]
#define H rgbsrc3[index]

#include "fill_dst_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    dst += 1;

    _nassert((int)dst % 2 == 0);
    _nassert((int)(dst-1) % 8 == 0);
    // upper right pixel

#define E rgbsrc2[index]
#define A rgbsrc1[index+1]
#define B rgbsrc2[index+1]
#define D rgbsrc1[index]
#define F rgbsrc3[index]
#define H rgbsrc2[index-1]

#include "fill_dst_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    dst += 2*width;

    _nassert((int)dst % 2 == 0);
    _nassert((int)(dst-1) % 8 == 0);
    // lower right pixel

#define E rgbsrc2[index]
#define A rgbsrc3[index+1]
#define B rgbsrc3[index]
#define D rgbsrc2[index+1]
#define F rgbsrc2[index-1]
#define H rgbsrc1[index]

#include "fill_dst_proc_body.h"

    blendTable += 20;
    blendValues += MAX_WIDTH;
    dst -= 1;

    _nassert((int)dst % 8 == 0);
    // lower left pixel

#define E rgbsrc2[index]
#define A rgbsrc3[index-1]
#define B rgbsrc2[index-1]
#define D rgbsrc3[index]
#define F rgbsrc1[index]
#define H rgbsrc2[index+1]

#include "fill_dst_proc_body.h"
}

void _hq2x_8_16(const uint8_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height, uint32_t *hqTable, uint16_t *rgbpalette)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);
    _nassert(srcstride >= width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);
    _nassert(height >= 4);
    _nassert((int)hqTable % 8 == 0);
    _nassert((int)rgbpalette % 8 == 0);

    // hqTable and rgbpalette should be in L1D SRAM

    uint8_t * restrict tmpsrc1;
    uint8_t * restrict tmpsrc2;
    uint8_t * restrict tmpsrc3;
    uint8_t * restrict tmpsrc4;

    uint16_t * restrict rgbsrc1;
    uint16_t * restrict rgbsrc2;
    uint16_t * restrict rgbsrc3;
    uint16_t * restrict rgbsrc4;

    uint32_t * restrict yuvsrc1;
    uint32_t * restrict yuvsrc2;
    uint32_t * restrict yuvsrc3;
    uint32_t * restrict yuvsrc4;

    uint8_t * restrict patternsrc1;

    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    uint8_t * restrict hqTable2;

    uint64_t * restrict rgbyuvpalette;

    uint16_t * restrict blendTable;

    tmpsrc1 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP); // tmpsrc1 is aligned to 8

    // start copying first line to tmpsrc1
    syscalls.qdma_copy1d(QDMA_0, tmpsrc1, src, width);

    tmpsrc2 = (uint8_t *) ( ((uintptr_t) tmpsrc1) + width); // tmpsrc2 is aligned to 8
    tmpsrc3 = (uint8_t *) ( ((uintptr_t) tmpsrc2) + width); // tmpsrc3 is aligned to 8
    tmpsrc4 = (uint8_t *) ( ((uintptr_t) tmpsrc3) + width); // tmpsrc4 is aligned to 8

    rgbsrc1 = (uint16_t *) ( ((uintptr_t) tmpsrc4) +     width + 8); // rgbsrc1 is aligned to 8 and has place for rgbsrc1[-1] and rgbsrc1[width]
    rgbsrc2 = (uint16_t *) ( ((uintptr_t) rgbsrc1) + 2 * width + 8); // rgbsrc2 is aligned to 8 and has place for rgbsrc2[-1] and rgbsrc2[width]
    rgbsrc3 = (uint16_t *) ( ((uintptr_t) rgbsrc2) + 2 * width + 8); // rgbsrc3 is aligned to 8 and has place for rgbsrc3[-1] and rgbsrc3[width]
    rgbsrc4 = (uint16_t *) ( ((uintptr_t) rgbsrc3) + 2 * width + 8); // rgbsrc4 is aligned to 8 and has place for rgbsrc4[-1] and rgbsrc4[width]

    yuvsrc1 = (uint32_t *) ( ((uintptr_t) rgbsrc4) + 2 * width + 8); // yuvsrc1 is aligned to 8 and has place for yuvsrc1[-1] and yuvsrc1[width]
    yuvsrc2 = (uint32_t *) ( ((uintptr_t) yuvsrc1) + 4 * width + 8); // yuvsrc2 is aligned to 8 and has place for yuvsrc2[-1] and yuvsrc2[width]
    yuvsrc3 = (uint32_t *) ( ((uintptr_t) yuvsrc2) + 4 * width + 8); // yuvsrc3 is aligned to 8 and has place for yuvsrc3[-1] and yuvsrc3[width]
    yuvsrc4 = (uint32_t *) ( ((uintptr_t) yuvsrc3) + 4 * width + 8); // yuvsrc4 is aligned to 8 and has place for yuvsrc4[-1] and yuvsrc4[width]

    patternsrc1 = (uint8_t *) ( ((uintptr_t) yuvsrc4) + 4 * width + 8); // patternsrc1 is aligned to 8 and has place for patternsrc1[-1] and patternsrc1[width]

    tmpdst1 = (uint16_t *) ( ((uintptr_t) patternsrc1) + width + 8); // tmpdst1 is aligned to 8
    tmpdst2 = (uint16_t *) ( ((uintptr_t) tmpdst1    ) + 8 * width); // tmpdst2 is aligned to 8

    hqTable2 = (uint8_t *) ( ((uintptr_t) tmpdst2) + 8 * width); // hqTable2 is aligned to 8

    rgbyuvpalette = (uint64_t *) ( ((uintptr_t) hqTable2) + 16); // rgbyuvpalette is aligned to 8

    blendTable = (uint16_t *) ( ((uintptr_t) rgbyuvpalette) + 8 * 256); // blendTable is aligned to 8

    hqTable2[ 0] =  4; hqTable2[ 1] = 0;
    hqTable2[ 2] = 10; hqTable2[ 3] = 0;
    hqTable2[ 4] = 11; hqTable2[ 5] = 0;
    hqTable2[ 6] =  4; hqTable2[ 7] = 1;
    hqTable2[ 8] =  9; hqTable2[ 9] = 1;
    hqTable2[10] = 10; hqTable2[11] = 1;

    prepare_yuv_palette(rgbpalette, rgbyuvpalette);

    // wait for copying first line to tmpsrc1
    syscalls.qdma_wait(QDMA_0);

    // start copying second line to tmpsrc2
    syscalls.qdma_copy1d(QDMA_0, tmpsrc2, src + srcstride, width);
    src += 2*srcstride;

    // convert first line to yuv and rgb
    convert_yuv_rgb_8(tmpsrc1, yuvsrc1, rgbsrc1, rgbyuvpalette, width);

    // wait for copying second line to tmpsrc2
    syscalls.qdma_wait(QDMA_0);

    // start copying third line to tmpsrc3
    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src, width);
    src += srcstride;

    // convert second line to yuv and rgb
    convert_yuv_rgb_8(tmpsrc2, yuvsrc2, rgbsrc2, rgbyuvpalette, width);

    // calculate pattern for first line
    calculate_pattern(yuvsrc1, yuvsrc1, yuvsrc2, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(rgbsrc1, rgbsrc1, rgbsrc2, tmpdst1, blendTable, width);

    // start copying first two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    dst += dststride;

    for (height -= 3; height != 0; height--)
    {
        // wait for copying "third" line to tmpsrc3
        syscalls.qdma_wait(QDMA_0);

        // convert "third" line to yuv and rgb
        convert_yuv_rgb_8(tmpsrc3, yuvsrc3, rgbsrc3, rgbyuvpalette, width);

        // calculate pattern for "second" line
        calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

        // calculate and process blend tables
        fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

        // start copying "fourth" line to tmpsrc4
        syscalls.qdma_copy1d(QDMA_0, tmpsrc4, src, width);
        src += srcstride;

        fill_dst(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2, blendTable, width);

        // wait for copying "first" two destination lines from tmpdst1
        syscalls.qdma_wait(QDMA_1);

        // start copying "second" two destination lines from tmpdst2
        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
        dst += dststride;

        // "rename" buffers
        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint8_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc;
        }
        {
            uint16_t * restrict rgbsrc;
            rgbsrc = rgbsrc1;
            rgbsrc1 = rgbsrc2;
            rgbsrc2 = rgbsrc3;
            rgbsrc3 = rgbsrc4;
            rgbsrc4 = rgbsrc;
        }
        {
            uint32_t * restrict yuvsrc;
            yuvsrc = yuvsrc1;
            yuvsrc1 = yuvsrc2;
            yuvsrc2 = yuvsrc3;
            yuvsrc3 = yuvsrc4;
            yuvsrc4 = yuvsrc;
        }
    }

    // last-1 line:

    // wait for copying last line to tmpsrc3
    syscalls.qdma_wait(QDMA_0);

    // convert last line to yuv and rgb
    convert_yuv_rgb_8(tmpsrc3, yuvsrc3, rgbsrc3, rgbyuvpalette, width);

    // calculate pattern for last line
    calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2, blendTable, width);

    // wait for copying last-2 two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_1);

    // start copying last-1 two destination lines from tmpdst2
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
    dst += dststride;

    // last line:

    // calculate pattern for last line
    calculate_pattern(yuvsrc2, yuvsrc3, yuvsrc3, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(rgbsrc2, rgbsrc3, rgbsrc3, tmpdst1, blendTable, width);

    // start copying last two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    //dst += dststride;

    // wait for copying last-1 two destination lines from tmpdst2
    syscalls.qdma_wait(QDMA_1);

    // wait for copying last two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_0);
}

void _hq2x_16_16(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height, uint32_t *hqTable)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);
    _nassert(height >= 4);
    _nassert((int)hqTable % 8 == 0);

    // hqTable should be in L1D SRAM

    uint16_t * restrict tmpsrc1;
    uint16_t * restrict tmpsrc2;
    uint16_t * restrict tmpsrc3;
    uint16_t * restrict tmpsrc4;

    uint32_t * restrict yuvsrc1;
    uint32_t * restrict yuvsrc2;
    uint32_t * restrict yuvsrc3;
    uint32_t * restrict yuvsrc4;

    uint8_t * restrict patternsrc1;

    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    uint8_t * restrict hqTable2;

    uint16_t * restrict blendTable;


    tmpsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8); // tmpsrc1 is aligned to 8 and has place for tmpsrc1[-1] and tmpsrc1[width]

    // start copying first line to tmpsrc1
    syscalls.qdma_copy1d(QDMA_0, tmpsrc1, src, 2*width);

    tmpsrc2 = (uint16_t *) ( ((uintptr_t) tmpsrc1) + 2 * width + 8); // tmpsrc2 is aligned to 8 and has place for tmpsrc2[-1] and tmpsrc2[width]
    tmpsrc3 = (uint16_t *) ( ((uintptr_t) tmpsrc2) + 2 * width + 8); // tmpsrc3 is aligned to 8 and has place for tmpsrc3[-1] and tmpsrc3[width]
    tmpsrc4 = (uint16_t *) ( ((uintptr_t) tmpsrc3) + 2 * width + 8); // tmpsrc4 is aligned to 8 and has place for tmpsrc4[-1] and tmpsrc4[width]

    yuvsrc1 = (uint32_t *) ( ((uintptr_t) tmpsrc4) + 2 * width + 8); // yuvsrc1 is aligned to 8 and has place for yuvsrc1[-1] and yuvsrc1[width]
    yuvsrc2 = (uint32_t *) ( ((uintptr_t) yuvsrc1) + 4 * width + 8); // yuvsrc2 is aligned to 8 and has place for yuvsrc2[-1] and yuvsrc2[width]
    yuvsrc3 = (uint32_t *) ( ((uintptr_t) yuvsrc2) + 4 * width + 8); // yuvsrc3 is aligned to 8 and has place for yuvsrc3[-1] and yuvsrc3[width]
    yuvsrc4 = (uint32_t *) ( ((uintptr_t) yuvsrc3) + 4 * width + 8); // yuvsrc4 is aligned to 8 and has place for yuvsrc4[-1] and yuvsrc4[width]

    patternsrc1 = (uint8_t *) ( ((uintptr_t) yuvsrc4) + 4 * width + 8); // patternsrc1 is aligned to 8 and has place for patternsrc1[-1] and patternsrc1[width]

    tmpdst1 = (uint16_t *) ( ((uintptr_t) patternsrc1) + width + 8); // tmpdst1 is aligned to 8
    tmpdst2 = (uint16_t *) ( ((uintptr_t) tmpdst1    ) + 8 * width); // tmpdst2 is aligned to 8

    hqTable2 = (uint8_t *) ( ((uintptr_t) tmpdst2) + 8 * width); // hqTable2 is aligned to 8

    blendTable = (uint16_t *) ( ((uintptr_t) hqTable2) + 16); // blendTable is aligned to 8

    hqTable2[ 0] =  4; hqTable2[ 1] = 0;
    hqTable2[ 2] = 10; hqTable2[ 3] = 0;
    hqTable2[ 4] = 11; hqTable2[ 5] = 0;
    hqTable2[ 6] =  4; hqTable2[ 7] = 1;
    hqTable2[ 8] =  9; hqTable2[ 9] = 1;
    hqTable2[10] = 10; hqTable2[11] = 1;

    // wait for copying first line to tmpsrc1
    syscalls.qdma_wait(QDMA_0);

    // start copying second line to tmpsrc2
    syscalls.qdma_copy1d(QDMA_0, tmpsrc2, src + srcstride/2, 2*width);
    src += srcstride;

    // convert first line to yuv
    convert_yuv_16(tmpsrc1, yuvsrc1, width);

    // wait for copying second line to tmpsrc2
    syscalls.qdma_wait(QDMA_0);

    // start copying third line to tmpsrc3
    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src, 2*width);
    src += srcstride/2;

    // convert second line to yuv
    convert_yuv_16(tmpsrc2, yuvsrc2, width);

    // calculate pattern for first line
    calculate_pattern(yuvsrc1, yuvsrc1, yuvsrc2, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(tmpsrc1, tmpsrc1, tmpsrc2, tmpdst1, blendTable, width);

    // start copying first two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    dst += dststride;

    for (height -= 3; height != 0; height--)
    {
        // wait for copying "third" line to tmpsrc3
        syscalls.qdma_wait(QDMA_0);

        // convert "third" line to yuv
        convert_yuv_16(tmpsrc3, yuvsrc3, width);

        // calculate pattern for "second" line
        calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

        // calculate and process blend tables
        fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

        // start copying "fourth" line to tmpsrc4
        syscalls.qdma_copy1d(QDMA_0, tmpsrc4, src, 2*width);
        src += srcstride/2;

        fill_dst(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2, blendTable, width);

        // wait for copying "first" two destination lines from tmpdst1
        syscalls.qdma_wait(QDMA_1);

        // start copying "second" two destination lines from tmpdst2
        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
        dst += dststride;

        // "rename" buffers
        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint16_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc;
        }
        {
            uint32_t * restrict yuvsrc;
            yuvsrc = yuvsrc1;
            yuvsrc1 = yuvsrc2;
            yuvsrc2 = yuvsrc3;
            yuvsrc3 = yuvsrc4;
            yuvsrc4 = yuvsrc;
        }
    }

    // last-1 line:

    // wait for copying last line to tmpsrc3
    syscalls.qdma_wait(QDMA_0);

    // convert last line to yuv
    convert_yuv_16(tmpsrc3, yuvsrc3, width);

    // calculate pattern for last-1 line
    calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2, blendTable, width);

    // wait for copying last-2 two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_1);

    // start copying last-1 two destination lines from tmpdst2
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
    dst += dststride;

    // last line:

    // calculate pattern for last line
    calculate_pattern(yuvsrc2, yuvsrc3, yuvsrc3, patternsrc1, width);

    // calculate and process blend tables
    fill_blendtable(patternsrc1, hqTable, hqTable2, blendTable, width);

    fill_dst(tmpsrc2, tmpsrc3, tmpsrc3, tmpdst1, blendTable, width);

    // start copying last two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    //dst += dststride;

    // wait for copying last-1 two destination lines from tmpdst2
    syscalls.qdma_wait(QDMA_1);

    // wait for copying last two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_0);
}

